import os
import numpy as np
import pandas as pd
from glob import glob
from gsitk.preprocess import pprocess_twitter, simple, Preprocesser
from sklearn.pipeline import Pipeline

## Read dataset
def read_tw_csv(path, split):
    df = pd.read_csv(path, sep='\t', header=None, names=['label', 'text_raw'])
    df['split'] = split
    return df

def read_dataset():
    tw_train = read_tw_csv('data/twitter/train/shuffled_dataset_LSTM.csv', 'train')
    tw_test = read_tw_csv('data/twitter/test/shuffled_dataset_LSTM.csv', 'test')
    dataset = pd.concat([tw_train, tw_test], axis=0).reset_index()

    pp_pipe = Pipeline([
        ('twitter', Preprocesser(pprocess_twitter)),
        ('simple', Preprocesser(simple)),
    ])

    dataset['text'] = pp_pipe.fit_transform(dataset['text_raw'])
    return dataset

## Study 1 dataset
def read_study1_dataset():
    df = pd.read_csv('data/new_dataset/moral_values_and_charitable_donation/study_1/dfProcessed',
            sep='\t', error_bad_lines=False)
    df = df[df['PostShare']=='post']
    df = df[df['Action']=='True']

    floats = ['HarmVirtue', 'HarmVice', 'FairnessVirtue', 'FairnessVice','IngroupVirtue', 'IngroupVice',
     'AuthorityVirtue', 'AuthorityVice', 'PurityVirtue', 'PurityVice', 'Morality']
    df[floats] = df[floats].astype(float)

    morality_mean = df['Morality'].mean()

    morals = ['harm', 'fairness', 'ingroup', 'authority', 'purity']
    for moral in morals:
        df[moral] = df[[col for col in floats if col.lower().startswith(moral)]].sum(axis=1)

    df['moral'] = df[morals].idxmax(axis=1)[df['Morality'] > morality_mean]
    df['moral'] = df['moral'].apply(lambda x: x if x in morals else 'non-moral')
    df['moral'] = df['moral'].apply(lambda x: x if x != 'ingroup' else 'loyalty')
    df['moral'] = df['moral'].apply(lambda x: x if x != 'harm' else 'care')
    df['text_raw'] = df['Tweet']
    df['label'] = df['moral']
    dataset = df[['text_raw', 'label']]

    pp_pipe = Pipeline([
        ('twitter', Preprocesser(pprocess_twitter)),
        ('simple', Preprocesser(simple)),
    ])

    dataset['text'] = pp_pipe.fit_transform(dataset['text_raw'])
    return dataset

    return df

## Read Lexicon
moral_label = {'fairness': 'FC', 'loyalty': 'LB', 'care': 'CH', 'purity': 'PD', 'authority': 'AS', 'non-moral': 'NM'}

def read_moral_lex():
    annotations_path = os.path.join(os.path.dirname(__file__), 'annotations')
    morals = [os.path.splitext(os.path.basename(file))[0] for file in glob(os.path.join(annotations_path, '*.tsv'))]

    moral_df = dict()
    moral_dict = dict()

    for moral in morals:
        tsv_path = os.path.join(annotations_path, "{}.tsv".format(moral))
        df = pd.read_csv(tsv_path, sep='\t')
        moral_df[moral] = df
        moral_dict[moral] = df['LEMMA'].values
    return moral_df, moral_dict



import numpy as np
import pandas as pd
import data 

moral_df, moral_dict = data.read_moral_lex()

moral_lex = dict()
for moral, df in moral_df.items():
    moral_lex[moral] = df.set_index('LEMMA')

moral_list = list(moral_lex.keys())

def moral_value(word, moral):
    try:
        v = moral_lex[moral].loc[word]['EXPRESSED_MORAL']
    except KeyError:
        v = -1
    if isinstance(v, pd.Series):
        return v.values[0]
    return v

def bucketize(x):
    if x == -1 :
        return np.array([0, 0])
    if x > 5:
        return np.array([0, 1])
    elif x <= 5:
        return np.array([1, 0])
    
def form_word_vector(word, bucketize_=None):
    v = []
    for moral in moral_lex.keys():
        v_m = moral_value(word, moral)
        if bucketize_ is not None:
            v_m = bucketize_(v_m)
        v.append(v_m)

    if bucketize_ is not None:
        return np.concatenate(v)

    v = np.array(v)
    return v
    
def form_text_vector(text, model='count'):
    '''
    model: count of freq
    '''
    if model == 'count':
        bucketize_ = bucketize
    elif model == 'freq':
        bucketize_ = None

    v = []
    for word in text:
        v.append(form_word_vector(word, bucketize_=bucketize_))
    if model == 'count':
        return np.sum(v, axis=0)
    elif model == 'freq':
        return np.concatenate(
            (
                np.average(v, axis=0),
                np.std(v, axis=0),
                np.median(v, axis=0),
                np.max(v, axis=0),
        ))




from moralstrength import lexicon_use
from moralstrength.estimators import estimate,models
import spacy
import en_core_web_sm

nlp = en_core_web_sm.load()
moral_options_lexicon = ['care', 'fairness', 'loyalty', 'authority', 'purity']
moral_options_predictions = moral_options_lexicon + ['non-moral']


def word_moral_value(word,moral):
    """Returns the association strength between word and moral trait,
       as rated by annotators. Value ranges from 1 to 9.
       1: words closely associated to harm, cheating, betrayal, subversion, degradation
       9: words closely associated to care, fairness, loyalty, authority, sanctity
       If the word is not in the lexicon of that moral trait, returns -1.
       For a list of available traits, get_available_lexicon_traits()
       """

    if moral not in moral_options_lexicon:
        raise ValueError('Invalid moral trait "{}" specified. Valid traits are: {}'.format(moral,moral_options_lexicon))
    return lexicon_use.moral_value(word=word, moral=moral)


def word_moral_values(word):
    """Returns a dict that gives the association strength between word and every
       moral trait, as rated by annotators. Value ranges from 1 to 9.
       1: words closely associated to harm, cheating, betrayal, subversion, degradation
       9: words closely associated to care, fairness, loyalty, authority, purity/sanctity
       If the word is not in the lexicon of that moral trait, returns -1."""

    return {moral: lexicon_use.moral_value(word=word, moral=moral) for moral in moral_options_lexicon}


def string_average_moral(text,moral):
    """Returns the average of the annotations for the words in the sentence (for one moral).
       If no word is recognized/found in the lexicon, returns -1.
       Words are lemmatized using spacy.
       """
    sum = 0
    recognized_words_no = 0
    for token in nlp(text):
        lemma = token.lemma_
        value = word_moral_value(lemma,moral)
        #print("lemma: {}".format(lemma))
        if value>-1:
            sum += value
            recognized_words_no += 1
    if recognized_words_no == 0:
        return -1
    else:
        return sum/recognized_words_no



def get_available_models():
    """Returns a list of available models for predicting texts.
       Short explanation of names:
       unigram: simple unigram-based model
       count: number of words that are rated as closer to moral extremes
       freq: distribution of moral ratings across the text
       simon: SIMilarity-based sentiment projectiON
       or a combination of these.
       For a comprehensive explanation of what each model does and how it performs on
       different datasets, see https://arxiv.org/abs/1904.08314
       (published at Knowledge-Based Systems)."""

    return models

def get_available_lexicon_traits():
    """Returns a list of traits that were annotated and can be queried
       by word_moral_value().
       care: Care/Harm
       fairness: Fairness/Cheating
       loyalty: Loyalty/Betrayal
       authority: Authority/Subversion
       purity: Purity/Degradation
       """

    return moral_options_lexicon

def get_available_prediction_traits():
    """Returns a list of traits that can be predicted by string_moral_value()
       or file_moral_value().
       care: Care/Harm
       fairness: Fairness/Cheating
       loyalty: Loyalty/Betrayal
       authority: Authority/Subversion
       purity: Purity/Degradation
       non-moral: Tweet/text is non-moral
       """

    return moral_options_predictions

def string_moral_value(text,moral,model='unigram+freq'):
    """Returns the estimated probability that the text is relevant to either a vice or
       virtue of the corresponding moral trait.
       The default model is unigram+freq, the best performing (on average) across all
       dataset, according to our work.
       For a list of available models, see get_available_models().
       For a list of traits, get_available_prediction_traits().
    """

    if model not in models:
        raise ValueError('Invalid model "{}" specified. Valid models are: {}'.format(model,models))
    if moral not in moral_options_predictions:
        raise ValueError('Invalid moral trait "{}" specified. Valid traits are: {}'.format(moral,moral_options_predictions))
    return estimate(text, moral=moral, model=model)[0]

def string_moral_values(text,model='unigram+freq'):
    """Returns the estimated probability that the text is relevant to vices or virtues
       of all moral traits, as a dict.
       The default model is unigram+freq, the best performing (on average) across all
       dataset, according to our work.
       For a list of available models, see get_available_models().
       For a list of traits, get_available_prediction_traits().
    """

    if model not in models:
        raise ValueError('Invalid model "{}" specified. Valid models are: {}'.format(moral,models))
    return {moral: estimate(text=text, moral=moral, model=model)[0] for moral in moral_options_predictions}

# def file_moral_value(filename,trait,model):
#   if model not in models:
#       raise ValueError('Invalid model "{}" specified. Valid models are: {}'.format(model,models))
#   if moral not in moral_options_predictions:
#       raise ValueError('Invalid moral trait "{}" specified. Valid traits are: {}'.format(moral,moral_options_predictions))
#
#   return
#
# def file_moral_values(filename,model):
#   if model not in models:
#       raise ValueError('Invalid model "{}" specified. Valid models are: {}'.format(model,models))
#   return



import setuptools

VERSION = '0.1.3'

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='moralstrength',
     packages=['moralstrength'],
     version=VERSION,
     #scripts=['moralstrength', 'data', 'estimators', 'lexicon_use', 'moralstrengthdict'] ,
     author="Oscar Araque, Lorenzo Gatti and Kyriaki Kalimeri",
     author_email="o.araque@upm.es",
     description="A package to predict the Moral Foundations for a tweet or text",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/oaraque/moral-foundations/",
     download_url='https://github.com/gsi-upm/gsitk/tarball/{}'.format(VERSION),
     license='LGPLv3',
     classifiers=[
     	 "Intended Audience :: Science/Research",
     	 "Natural Language :: English",
     	 "Topic :: Text Processing :: Linguistic",
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
         "Operating System :: OS Independent",
     ],
     keywords=['moral foundations', 'NLP', 'moralstrength', 'machine learning'],
	 install_requires=[
   		'gsitk',
		'spacy',
		'numpy',
		'pandas',
		'scikit-learn==0.20.0',
	],
    include_package_data=True,
 )



import os
import numpy as np
import pickle
from sklearn.pipeline import Pipeline
from gsitk.preprocess import pprocess_twitter, simple, Preprocesser

from moralstrength.lexicon_use import form_text_vector

moral_options = ('care', 'fairness', 'loyalty', 'authority', 'purity', 'non-moral')
models = (
    'simon', 'unigram', 'count', 'freq',
    'simon+count', 'simon+freq', 'simon+count+freq',
    'unigram+count', 'unigram+freq', 'unigram+count+freq',
    'simon+unigram+count', 'simon+unigram+freq', 'simon+unigram+count+freq',
)

sorter = {'simon': 0, 'unigram': 1, 'count': 2, 'freq': 3}

pp_pipe = Pipeline([
        ('twitter', Preprocesser(pprocess_twitter)),
        ('simple', Preprocesser(simple)),
])

def generate_model_name(names):
    names = sorted(zip(names, [sorter[i] for i in names]), key=lambda x: x[1])
    return '+'.join([n[0] for n in names])

def pck_name(model_name, moral, transformer=False, transformer_name='unigram'):
    if not transformer:
        return "{}_{}_lr".format(model_name, moral)
    if transformer_name == 'unigram':
        return "ngram"
    elif transformer_name == 'simon':
        return "simon_{}".format(moral)

def load_model(model_name):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    model_path = 'export/{}.pck'.format(model_name)
    full_path = os.path.join(dir_path, model_path)
    with open(full_path, 'rb') as f:
        m = pickle.load(f)
    return m

def load_models(estimators, transformers, moral):
    model_name = generate_model_name(estimators)
    lr = load_model(pck_name(model_name, moral))

    trans = {}
    for transformer in transformers:
        if transformer == 'unigram' or transformer == 'simon':
            name_tmp = pck_name(
                transformer, moral,
                transformer=True, transformer_name=transformer
                )
            trans[transformer] = load_model(name_tmp)
        else:
            trans[transformer] = None

    return lr, trans

def get_estimators(estimator_name):
    return estimator_name.split('+')

def select_processes(process_names, moral):
    estimators = get_estimators(process_names)
    transformers = []
    for estimator in estimators:
        transformers.append(estimator)
        #if estimator == 'simon':
        #    transformers.append('simon')
        #elif estimator == 'unigram':
        #    transformers.append('unigram')
    return estimators, transformers



def estimate(text, moral, model='count', from_file=False):
    if not from_file:
        text_processed = pp_pipe.transform([text])
    else:
        with open('lines.txt', 'r') as f:
            lines = f.readlines()
        lines = [line.strip() for line in lines]
        text_processed = pp_pipe.transform(lines)

    estimators, transformers = select_processes(model, moral)
    lr, transformers = load_models(estimators, transformers, moral)

    X = []
    for transformer in transformers.keys():
        if transformer == 'unigram':
            X_tmp = transformers[transformer].transform([' '.join(t) for t in text_processed])
            X_tmp = X_tmp.toarray()
            X.append(X_tmp)
        elif transformer == 'simon':
            X_tmp = transformers[transformer].transform(text_processed)
            X.append(X_tmp)
        else: 
            X_tmp = [form_text_vector(t, model=transformer) for t in text_processed]    
            X.append(X_tmp)
    X = np.hstack(X)

    return lr.predict_proba(X)[:, 1]
